@@ -14,7 +14,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * SCCS: @(#) tclStringObj.c 1.31 97/10/30 13:56:35
+ * RCS: @(#) $Id: tclStringObj.c,v 1.6 1999/05/07 20:07:35 stanton Exp $
  */
 
 #include "tclInt.h"
@@ -74,9 +74,9 @@ Tcl_ObjType tclStringType = {
 
 Tcl_Obj *
 Tcl_NewStringObj(bytes, length)
-    register CONST char *bytes;	/* Points to the first of the length bytes
+    CONST char *bytes;		/* Points to the first of the length bytes
 				 * used to initialize the new object. */
-    register int length;	/* The number of bytes to copy from "bytes"
+    int length;			/* The number of bytes to copy from "bytes"
 				 * when initializing the new object. If 
 				 * negative, use bytes up to the first
 				 * NULL byte. */
@@ -88,9 +88,9 @@ Tcl_NewStringObj(bytes, length)
 
 Tcl_Obj *
 Tcl_NewStringObj(bytes, length)
-    register CONST char *bytes;	/* Points to the first of the length bytes
+    CONST char *bytes;		/* Points to the first of the length bytes
 				 * used to initialize the new object. */
-    register int length;	/* The number of bytes to copy from "bytes"
+    int length;			/* The number of bytes to copy from "bytes"
 				 * when initializing the new object. If 
 				 * negative, use bytes up to the first
 				 * NULL byte. */
@@ -140,9 +140,9 @@ Tcl_NewStringObj(bytes, length)
 
 Tcl_Obj *
 Tcl_DbNewStringObj(bytes, length, file, line)
-    register CONST char *bytes;	/* Points to the first of the length bytes
+    CONST char *bytes;		/* Points to the first of the length bytes
 				 * used to initialize the new object. */
-    register int length;	/* The number of bytes to copy from "bytes"
+    int length;			/* The number of bytes to copy from "bytes"
 				 * when initializing the new object. If 
 				 * negative, use bytes up to the first
 				 * NULL byte. */
@@ -165,7 +165,7 @@ Tcl_DbNewStringObj(bytes, length, file, line)
 
 Tcl_Obj *
 Tcl_DbNewStringObj(bytes, length, file, line)
-    register CONST char *bytes;	/* Points to the first of the length bytes
+    CONST char *bytes;		/* Points to the first of the length bytes
 				 * used to initialize the new object. */
     register int length;	/* The number of bytes to copy from "bytes"
 				 * when initializing the new object. If 
@@ -222,12 +222,6 @@ Tcl_SetStringObj(objPtr, bytes, length)
 	panic("Tcl_SetStringObj called with shared object");
     }
 
-    Tcl_InvalidateStringRep(objPtr);
-    if (length < 0) {
-	length = strlen(bytes);
-    }
-    TclInitStringRep(objPtr, bytes, length);
-        
     /*
      * Set the type to NULL and free any internal rep for the old type.
      */
@@ -236,6 +230,12 @@ Tcl_SetStringObj(objPtr, bytes, length)
 	oldTypePtr->freeIntRepProc(objPtr);
     }
     objPtr->typePtr = NULL;
+
+    Tcl_InvalidateStringRep(objPtr);
+    if (length < 0) {
+	length = (bytes? strlen(bytes) : 0);
+    }
+    TclInitStringRep(objPtr, bytes, length);
 }
 
 /*
@@ -335,7 +335,7 @@ Tcl_AppendToObj(objPtr, bytes, length)
 	ConvertToStringType(objPtr);
     }
     if (length < 0) {
-	length = strlen(bytes);
+	length = (bytes? strlen(bytes) : 0);
     }
     if (length == 0) {
 	return;
@@ -363,7 +363,36 @@ Tcl_AppendToObj(objPtr, bytes, length)
 /*
  *----------------------------------------------------------------------
  *
- * Tcl_AppendStringsToObj --
+ * Tcl_AppendObjToObj --
+ *
+ *	This procedure appends the string rep of one object to another.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	The string rep of appendObjPtr is appended to the string 
+ *	representation of objPtr.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+Tcl_AppendObjToObj(objPtr, appendObjPtr)
+    Tcl_Obj *objPtr;		/* Points to the object to append to. */
+    Tcl_Obj *appendObjPtr;	/* Object to append. */
+{
+    int length;
+    char *stringRep;
+
+    stringRep = Tcl_GetStringFromObj(appendObjPtr, &length);
+    Tcl_AppendToObj(objPtr, stringRep, length);
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Tcl_AppendStringsToObjVA --
  *
  *	This procedure appends one or more null-terminated strings
  *	to an object.
@@ -379,14 +408,14 @@ Tcl_AppendToObj(objPtr, bytes, length)
  */
 
 void
-Tcl_AppendStringsToObj TCL_VARARGS_DEF(Tcl_Obj *,arg1)
+Tcl_AppendStringsToObjVA (objPtr, argList)
+    Tcl_Obj *objPtr;		/* Points to the object to append to. */
+    va_list argList;		/* Variable argument list. */
 {
-    va_list argList;
-    register Tcl_Obj *objPtr;
+    va_list tmpArgList;
     int newLength, oldLength;
     register char *string, *dst;
 
-    objPtr = (Tcl_Obj *) TCL_VARARGS_START(Tcl_Obj *,arg1,argList);
     if (Tcl_IsShared(objPtr)) {
 	panic("Tcl_AppendStringsToObj called with shared object");
     }
@@ -397,12 +426,14 @@ Tcl_AppendStringsToObj TCL_VARARGS_DEF(Tcl_Obj *,arg1)
     /*
      * Figure out how much space is needed for all the strings, and
      * expand the string representation if it isn't big enough. If no
-     * bytes would be appended, just return.
+     * bytes would be appended, just return.  Note that on some platforms
+     * (notably OS/390) the argList is an array so we need to use memcpy.
      */
 
+    memcpy ((VOID *) &tmpArgList, (VOID *) &argList, sizeof (tmpArgList));
     newLength = oldLength = objPtr->length;
     while (1) {
-	string = va_arg(argList, char *);
+	string = va_arg(tmpArgList, char *);
 	if (string == NULL) {
 	    break;
 	}
@@ -430,7 +461,6 @@ Tcl_AppendStringsToObj TCL_VARARGS_DEF(Tcl_Obj *,arg1)
      * strings to the object.
      */
 
-    TCL_VARARGS_START(Tcl_Obj *,arg1,argList);
     dst = objPtr->bytes + oldLength;
     while (1) {
 	string = va_arg(argList, char *);
@@ -455,6 +485,34 @@ Tcl_AppendStringsToObj TCL_VARARGS_DEF(Tcl_Obj *,arg1)
 	*dst = 0;
     }
     objPtr->length = newLength;
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Tcl_AppendStringsToObj --
+ *
+ *	This procedure appends one or more null-terminated strings
+ *	to an object.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	The contents of all the string arguments are appended to the
+ *	string representation of objPtr.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+Tcl_AppendStringsToObj TCL_VARARGS_DEF(Tcl_Obj *,arg1)
+{
+    register Tcl_Obj *objPtr;
+    va_list argList;
+
+    objPtr = TCL_VARARGS_START(Tcl_Obj *,arg1,argList);
+    Tcl_AppendStringsToObjVA(objPtr, argList);
     va_end(argList);
 }
 
